unit paddapp;
{holds the main functions of the application itself}

{{$DEFINE DEBUG}
{$DEFINE FINAL}

interface

procedure initPADD;
procedure viewDeleted;
procedure saveSlack;
procedure freezeDeleted;

implementation

uses
  {$IFDEF DEBUG}
  ztimer,
  {$ENDIF}
  PADDvars,
  PADDfunctions,
  cmdlin,
  tfastwrite,
  support;

procedure initPADD;

const
  {$IFDEF DEBUG}
  dirsample:byte=3;
  {$ELSE}
  dirsample:byte=9;
  {$ENDIF}
  {RAM for 4K screen + 4K scratch + 32K cluster * 2 + FAT copy?}
  {memoryNeeds:longint=(128+4+4+(32*2)) * 1024;}

var
  s,fn:string;
  w,w2,clusval:word;
  b,b2,ab,abloop:byte;
  p:pointer;
  dirs:prootdirs;
  clusbuf:pointer;
  i:integer;
  l:longint;
  wp:^word;
  cp:^char;

  procedure CheckUserSpooked;
  begin
    writeln('If this bothers you, run CHKDSK, SCANDISK, or Norton Disk Doctor to see if');
    writeln('anything needs fixing (but is not a guarantee the DOS value will get fixed).');
    if not confirm('Type Y to continue, or any other key to exit to DOS: ')
      then abort(11,'User requested exit');
  end;

begin
  writeln(#13#10,progbanner);

  {allocate 128K sequentially for FAT copy, will free some later if not all needed}
  if maxavail<(128*1024) then abort(6,'Not enough RAM to allocate contiguously for FAT copy');
  {allocate working around TP's FAR pointer limitation}
  getmem(f1,segsize); getmem(f2,segsize); getmem(f3,segsize); getmem(f4,segsize);
  if  (seg(f4^)-seg(f3^)=(segsize div 16))
  and (seg(f3^)-seg(f2^)=(segsize div 16))
  and (seg(f2^)-seg(f1^)=(segsize div 16))
  then begin
    FATCopy:=f1;
    fillchar(f1^,segsize,$ff); fillchar(f2^,segsize,$ff);
    fillchar(f3^,segsize,$ff); fillchar(f4^,segsize,$ff);
    {populate lookup table used by FATVal}
    FATSegs[0]:=seg(f1^);
    FATSegs[1]:=seg(f3^);
  end else abort(7,'Unable to allocate 128K RAM contiguously');

  tfastinit(80,25);

  with DriveInfo do begin

    s:=non_flag_param(1);
    if (s[2]<>':') or (paramcount=0) or is_param('?') or is_param('h') then begin
      PrintCmdLineHelp;
      halt(1);
    end;
    Letter:=upcase(s[1]);
    if is_param('v') then verbose:=true;
    if is_param('y') then skipprompts:=true;
    if is_param('s') then tfPreventSnow:=true;
    if is_param('a') then aggressiveFsearch:=true;
    {if is_param('f') or is_param('k') then abort(1,'User selected unimplemented feature.');}

    writeln(#13#10'Before we begin, let''s perform some sanity checks.');
    write('DOS is checking cluster utilization, please wait up to 30 seconds... ');
    {Check for full drive, since we are useless on those}
    asm
      mov     ah,36h              {get free disk space}
      mov     dl,DriveInfo.Letter
      sub     dl,64               {a=1, z=26}
      int     21h
      mov     DriveInfo.NFC,bx
    end;
    writeln('done.');
    if NFC=0 then abort(2,'Drive appears full -- no deleted data to examine!');
    {gather drive stats and confirm with user}
    asm
      push    ds
      mov     ah,1ch
      mov     dl,DriveInfo.Letter
      sub     dl,64       {a=1, z=26}
      int     21h
      mov     ah,ds:[bx]
      pop     ds
      mov     DriveInfo.MediaDescriptor,ah
      mov     DriveInfo.SPC,al
      mov     DriveInfo.BPS,cx
      mov     DriveInfo.NC,dx
    end;
    if (MediaDescriptor AND 128)<>128 then abort(2,'Media descriptor invalid');
    if (MediaDescriptor AND 4)=4 then Removable:=true else Removable:=false;
    write('Drive ',Letter,': appears to be a '); if Removable then write('removable ');
    case MediaDescriptor of
      $F8:if removable then s:='1.44MB DSQD' else s:='Hard disk';
      $F9:s:='5" DSHD or 3" DSDD disk';
      $FA:s:='RAM disk';
      $FC:s:='5" SSDD disk';
      $FD:s:='5" DSDD disk';
      $FE:s:='5" SSDD 8-sector disk';
      $FF:s:='5" DSDD 8-sector disk';
    else
      s:='unknown disk (!?)';
    end;
    writeln(s,'...');
    if verbose then begin
      writeln('with the following characteristics:');
      writeln('   Bytes per sector: ',BPS);
      writeln('Sectors per cluster: ',SPC);
      writeln('   Clusters on disk: ',NC);
    end;
    TotalBytes:=longint(SPC)*longint(NC)*longint(BPS);
    FreeBytes:=longint(SPC)*longint(NFC)*longint(BPS);

    {determine FAT type based on Microsoft 1999 whitepaper}
    if NC < 4085
      then FAT:=FAT12
      else if NC < 65525
        then FAT:=FAT16
        else FAT:=FAT32;
    if fat=fat32 then abort(32,'No FAT32 support.  If you need it, contact the author.');
    case fat of
      FAT12:begin
        FATstring:='FAT12';
        EOC:=$ff8;
      end;
      FAT16:begin
        FATstring:='FAT16';
        EOC:=$fff8;
      end;
    end;

    writeln('...and holds a ',TotalBytes div 1024,' KiB ',FATstring,
            ' filesystem with ',FreeBytes div 1024,' KiB free remaining.');
    writeln('Reading disk and filesystem structures... ');
    p:=@DPT;
    asm
      push    ds
      les     di,p
      mov     ah,32h      {Get drive param table -- undocumented but CHKDSK uses it since 2.x}
      mov     dl,DriveInfo.Letter
      sub     dl,64       {a=1, z=26}
      int     21h
      mov     cx,DPTSize
      mov     si,bx
      rep     movsb       {copy into our own local var struct}
      pop     ds
    end;
    FATStartSec:=DPT.NSBFAT;
    FATNumSectors:=DPT.FATNumSecs;
    {limit FATNumSectors if fat12 filesystem, which means hi byte is likely junk}
    {If this is Compaq DOS 3.31, FNS cannot be over 256, so also truncate}
    if (fat=fat12) or (lo(DosVer)<=3) then FATNumSectors:=FATNumSectors AND $ff;
    if FatNumSectors=0 then FatNumSectors:=256; {adjust for 0=256}
    {$IFDEF DEBUG}
    writeln('NSBFAT: ',DPT.NSBFAT,' FATCOPIES: ',DPT.FATCOPIES,' FATnumSec: ',FATNumSectors);
    {$ENDIF}
    ROOTStartSec:=DPT.NSBFAT+(DPT.FATCOPIES*FATNumSectors);
    CLUSStartSec:=DPT.NFDS;
    { This works, but I try to trust DOS throughout the entire program
    so I'll use DOS's value.
    w:=(DPT.NRDE*32) div 512;
    if ((DPT.NRDE*32) mod 512)<>0 then inc(w);
    CLUSStartSec:=ROOTStartSec+w; }
    if verbose then begin
      writeln('There are ',DPT.FATCOPIES,' x ',FATNumSectors,'-sector FATs starting at sector ',FATStartSec);
      writeln('The Root directory starts at sector ',ROOTStartSec,' and contains ',DPT.NRDE,' entries');
      writeln('Data clusters start at sector ',CLUSStartSec);
    end;

    writeln('Here are the first few root directory entries; please check for consistency:'#13#10);
    if verbose then begin
      writeln('FILENAME.EXT <TYP> ATTRIBS   FILESIZE  Start  Modified   Created   LastAccess');
      writeln('------------ ----- -------- ---------- ----- ---------- ---------- ----------');
    end else begin
      writeln('FILENAME.EXT <TYP> ATTRIBS   FILESIZE  Start  Modified ');
      writeln('------------ ----- -------- ---------- ----- ----------');
    end;
    getmem(dirs,1024);
    fillchar(dirs^,1024,0);
    DOSAbsDiskRead(Letter,ROOTStartSec,2,dirs,gIOerr);
    if not verbose then inc(dirsample,7);
    for w:=0 to dirsample-1 do begin
      s[0]:=#12; s[9]:='.';
      for b:=0 to 7 do s[b+1]:=dirs^[w].name[b];
      for b:=0 to 2 do s[b+10]:=dirs^[w].ext[b];
      fn:=s; {save for later error checking}
      case ord(s[1]) of
        $00:write('UNUSED ENTRY');
        $E5:write('ERASED ENTRY');
        else begin
          if s[1]=chr($05) then s[1]:=chr($E5); {adjust for Kanji lead byte}
          write(s);
        end;
      end;
      ab:=dirs^[w].attr;
      if (ab and da_LFN)=da_LFN
        then write(' <LFN> ')
        else if (ab and da_DIR)=da_DIR
          then write(' <DIR> ')
          else if (ab and da_VOL)=da_VOL
            then write(' <VOL> ')
            else write('       ');

      s:='!?ADVSHR';
      {block out any flag that isn't actually bitset}
      asm
        lea si,s
        mov di,si
        mov ax,ss
        mov es,ax
        segss lodsb
        stosb
        mov cl,al
        xor ch,ch
        mov bl,ab
      @bits:
        segss lodsb
        shl bl,1
        jc @loopit
        mov al,'-'
      @loopit:
        stosb
        loop @bits
      end;
      write(s,' ');
      write(intpadded(dirs^[w].filesize,10,#32),' ');
      write(numfmt(dirs^[w].FirstClusLO),' ',dosfndate(dirs^[w].WriteDate));
      if verbose then write(' ',dosfndate(dirs^[w].CrtDate),' ',dosfndate(dirs^[w].LastAccDate));
      writeln;
      {check for invalid characters in filename}

      w2:=0;
      for b:=1 to 12 do begin
        {Microsoft whitepaper says these chars are not legal in any part of filename}
        if (b <> 9) {skip the '.'}
          then if byte(fn[b]) in [$1..$4,$6..$1f,$22,$2a..$2c,$2e,$2f,$3a..$3f,$5b..$5d,$7c]
            then inc(w2);
      end;
      if w2<>0 then begin
        writeln('That last filename appeared to contain illegal characters.');
        CheckUserSpooked;
      end;
    end;
    freemem(dirs,1024);
    if not confirm('Does everything look valid?')
      then abort(3,'Program malfunctioning and/or user lost confidence in program');

    if verbose then write('Caching FAT into RAM...');
    CacheFATFromDisk;
    if verbose then writeln('...done.');
    {free up some RAM if we don't need it for the copy}
    FATsizeb:=longint(FATNumSectors) * 512;
    if FATsizeb < (segsize*3) then freemem(f4,segsize);
    if FATsizeb < (segsize*2) then freemem(f3,segsize);
    if FATsizeb < segsize     then freemem(f2,segsize);

    {At this point we should have a copy of the FAT in memory at FATCopy^.
    As a test, seek to the very first free cluster and print contents}
    {first valid cluster is #2}

    {$IFDEF DEBUG}
    {seek to first free cluster and print its contents if debug}
    _pztimeron;
    FATval($10);
    _pztimeroff;
    _pztimerreport;
    writeln('Testing cluster mechanisms.  Searching for first free cluster...');
    getmem(clusbuf,512*SPC);
    for w:=2 to NC+1 do begin
      if FATVal(w)=$0000 then begin
        ReadCluster(w,clusbuf);
        break;
      end;
    end;
    if w=NC+1 then abort(7,'Couldn''t find a free cluster; program likely malfunctioning');
    {Throw first cluster onscreen}
    writeln('First free cluster is ',w,' and has the following printable contents:');
    cp:=clusbuf;
    s:='';
    for w:=0 to 512*SPC do begin
      if cp^ in [#32..#127] then s:=s+cp^;
      if w mod 80=0 then begin
        write(s);
        s:='';
      end;
      inc(word(cp));
    end;
    freemem(clusbuf,512*SPC);
    {$ENDIF}

    write(#13#10,'Counting free clusters, please wait... ');
    CFC:=FATValFree; writeln('done.');

    {Check for various errors}
    w:=0;
    if CFC <> NFC then begin
      writeln('DOS is incorrectly reporting the number of free clusters in the FAT.');
      writeln('DOS claims ',NFC,' but we counted ',CFC,' ('#241,NFC - CFC,') in the FAT itself.');
      inc(w);
    end;
    if fat=fat16 then begin
      if (FATVal(1) AND FAT16_ClnShutdown)<>FAT16_ClnShutdown then begin
        writeln('CleanShutdown bit missing; this volume may not have been unmounted cleanly.');
        inc(w);
      end;
      if (FATVal(1) AND FAT16_NoHardError)<>FAT16_NoHardError then begin
        writeln('NoHardError bit missing; this volume may have encountered disk I/O errors.');
        inc(w);
      end;
    end;
    if w<>0 then CheckUserSpooked;

  end; {with}
end;

procedure viewDeleted;

var
  ch:char;
  w,w2,w3,           {temp word}
  curclus,           {current cluster #}
  curfreeclus,       {current free cluster #}
  lastshown,         {last cluster buffer byte shown+1}
  curbufpos,         {current buffer display position}
  viewpagesize,      {display page size in bytes, used for paging}
  viewlinesize,      {display page size in bytes, used for paging}
  clussize:word;     {size of cluster in bytes}
  clusbuf:pointer;
  sdest:pointer;
  s:string;

const
  viewlines=23;
  l_toptemplate=' Unused Cluster xxxxx (xxxxx of xxxxx)  Viewing xxxxx - xxxxx  Display: ..... ';
  l_delclus=16;
  l_curfree=23;
  l_counted=32;
  l_startv=49;         {start of viewing area}
  l_endv=57;           {end of viewing area}
  l_viewm=74;


  jumpsize:word=256;   {how many clusters to longjump ahead/behind}

type
  str80=string[80];

var
  tempbuf:pointer;
  oldcurclus,oldcurfreeclus:word;
  foundDelData:boolean;

  procedure RepaintTopLine;
  var
    s1:str80;
  begin
    s1:=l_toptemplate;
    tfastwritexy(@s1,sdest,0,0,$70);
    with DriveInfo do begin
      s:=numfmt(curclus);     tfastwritexy(@s,sdest,l_delclus,0,$70);
      s:=numfmt(curfreeclus); tfastwritexy(@s,sdest,l_curfree,0,$70);
      s:=numfmt(CFC);         tfastwritexy(@s,sdest,l_counted,0,$70);
      s:=numfmt(curbufpos);   tfastwritexy(@s,sdest,l_startv,0,$70);
      case viewmode of
          rawview:s:='RAW  ';
        hexview16:s:='HEX16';
        hexview20:s:='HEX20';
        hexview24:s:='HEX24';
        printable:s:='ASCII';
      end;
      tfastwritexy(@s,sdest,l_viewm,0,$70);
    end;
  end;

  procedure RepaintBottomLine;
  var
    s1:str80;
  begin
    s1:='Keys: F1=Help  PgUp/PgDn/Home/End/'#24'/'#25'=navigate cluster  '#27'/'#26'=prev/next cluster ';
    tfastwritexy(@s1,sdest,0,24,$70);
  end;

  procedure ShowStatus(ss:str80;imp:boolean);
  {prints a status message directly to the screen.  If important, blink.}
  var
    s:str80;
    bf:byte;
  begin
    fillchar(s,81,#32); s:=ss; s[0]:=chr(80);
    if imp then bf:=$F0 else bf:=$70;
    tfastwritexy(@s,vidp,0,0,bf);
  end;

begin
  {$IFDEF DEBUG}
  sdest:=vidp;
  {$ELSE}
  getmem(sdest,80*25*2);
  {$ENDIF}
  with DriveInfo do begin

    clussize:=512*SPC;
    getmem(clusbuf,clussize);
    getmem(tempbuf,clussize);
    {init vars}
    curclus:=2;
    {find first free cluster}
    w:=FindNextFreeClus(curclus);
    if w<>0 then begin
      curclus:=w;
      curfreeclus:=1;
      ReadCluster(curclus,clusbuf);
    end else abort(10,'Couldn''t locate first free cluster');
    lastshown:=0; curbufpos:=0; viewpagesize:=0; viewlinesize:=0;

    RepaintTopLine;
    RepaintBottomLine;
    repeat
      {display what we have in the buffer, then get input}

      {display the cluster buffer and set the paging distances}
      RepaintTopLine;
      lastshown:=Showbuffer(clusbuf,sdest,clussize,curbufpos,viewmode);
      if viewpagesize=0 then begin
        viewpagesize:=lastshown;
        viewlinesize:=(viewpagesize div viewlines);
      end;
      s:=numfmt(curbufpos+lastshown-1); tfastwritexy(@s,sdest,l_endv,0,$70);
      {$IFNDEF DEBUG}
      tfastcopyscreen(sdest,vidp);
      {$ENDIF}

      ch:=readkeychar;
      case ch of
      'h':numhex:=not numhex;
      'v':begin
        {cycle through view modes}
        inc(byte(viewmode)); if byte(viewmode)>4 then byte(viewmode):=0;
        viewpagesize:=0; curbufpos:=0;
      end;
      'f':begin
        BIOSClrScr;
        FreezeDeleted;
        exit;
      end;
      #77:begin {right: advance to next free cluster}
        w:=FindNextFreeClus(curclus+1);
        if w<>0 then begin
          curclus:=w;
          inc(curfreeclus);
          ReadCluster(curclus,clusbuf);
          {reset viewpagesize and curbufpos}
          {b/c we won't know what they are until cluster is displayed}
          viewpagesize:=0; curbufpos:=0;
        end;
      end;
      #75:begin {left: advance to previous free cluster}
        w:=FindPrevFreeClus(curclus-1);
        if w<>0 then begin
          curclus:=w;
          dec(curfreeclus);
          ReadCluster(curclus,clusbuf);
          {reset viewpagesize and curbufpos}
          viewpagesize:=0; curbufpos:=0;
        end;
      end;
      #81:begin {pgdn: display next page of cluster}
        {don't advance display position past limit}
        if clussize>viewpagesize
          then if curbufpos+viewpagesize<clussize
            then inc(curbufpos,viewpagesize);
      end;
      #73:begin {pgup: display previous page of cluster}
        if clussize>viewpagesize
          then if curbufpos>viewpagesize
            then dec(curbufpos,viewpagesize)
            else curbufpos:=0;
      end;
      #80:begin {down: scroll down one line}
        if (clussize>viewpagesize) and (viewmode<>printable)
          then if curbufpos+viewlinesize<clussize
            then begin
              inc(curbufpos,viewlinesize);
              {prevent hex routines from trying to draw partial lines}
              if clussize-curbufpos<viewlinesize
                then curbufpos:=clussize-viewlinesize;
            end;
      end;
      #72:begin {up: scroll up one line}
        if (clussize>viewpagesize) and (viewmode<>printable)
          then if curbufpos>viewlinesize
            then dec(curbufpos,viewlinesize)
            else curbufpos:=0;
      end;
      #79:begin {end: }
        if clussize>viewpagesize
          then curbufpos:=clussize-viewpagesize;
      end;
      #71:begin {home: }
        curbufpos:=0;
      end;
      #54:begin {shift-right: advance to next cluster in bigger jumps}
        w3:=curclus;
        for w2:=0 to jumpsize-1 do begin
          w:=FindNextFreeClus(curclus+1);
          if w<>0 then begin
            curclus:=w;
            inc(curfreeclus);
            viewpagesize:=0; curbufpos:=0;
          end else break;
        end;
        if w3<>curclus
          then ReadCluster(curclus,clusbuf);
      end;
      #52:begin {shift-left: back to prev cluster in bigger jumps}
        w3:=curclus;
        for w2:=0 to jumpsize-1 do begin
          w:=FindPrevFreeClus(curclus-1);
          if w<>0 then begin
            curclus:=w;
            dec(curfreeclus);
            viewpagesize:=0; curbufpos:=0;
          end else break;
        end;
        if w3<>curclus
          then ReadCluster(curclus,clusbuf);
      end;
      #116:begin {ctrl-right: skip to next non-empty free cluster}
        oldcurclus:=curclus; oldcurfreeclus:=curfreeclus; foundDelData:=false;
        for w2:=curfreeclus to CFC-1 do begin
          {get next free cluster value}
          w:=FindNextFreeClus(curclus+1);
          {0=couldn't find one, <>0 found one}
          if w<>0 then begin
            {found next free cluster, update vars, make it current}
            inc(curfreeclus); viewpagesize:=0; curbufpos:=0; curclus:=w;
            {read it into the temp buffer so we can take a look at it}
            ShowStatus('Examining cluster '+numfmt(curclus)+' for deleted data...',false);
            ReadCluster(curclus,tempbuf);
            if not SameVal(tempbuf,clussize,$f6) then begin
              {found cluster where contents are not same or formatted value}
              move(tempbuf^,clusbuf^,clussize);
              RepaintTopLine;
              FoundDelData:=true;
              break;
            end;
          end;
        end;
        if not foundDelData then begin
          curclus:=oldcurclus; curfreeclus:=oldcurfreeclus;
          ShowStatus('Nothing interesting found past cluster '+numfmt(oldcurclus)+', returning to origin -- press a key',true);
          {wait for user to acknowledge}
          if keypressed then repeat readkeychar until not keypressed; readkeychar;
        end;
      end;
      #115:begin {ctrl-left: skip to previous non-empty free cluster}
      end;
      #59:begin {F1: help}
        PreserveDOSScreen;
        BIOSClrScr;
        PrintViewerHelp;
        if keypressed then repeat readkeychar until not keypressed; readkeychar;
        RestoreDOSScreen;
      end;
    end; {case}
    until ch=#27;

    freemem(tempbuf,clussize);
    freemem(clusbuf,clussize);

  end; {with DriveInfo}
  {$IFNDEF DEBUG}
  freemem(sdest,80*25*2);
  {$ENDIF}
  BIOSClrscr;
end;

procedure SaveSlack;
var
  totalslack,totalfiles:longint;
  s:string;
  f:file;
  clusbytes:word;
  clusbuf:pointer;

  Procedure IterateDIR(startclus:word;startsec:longint;numentries:word;level:byte;printonly:boolean);
  {startclus and level are 0 for root dir, nonzero otherwise}
  {todo: optimize so that only startclus needs to be passed}
  var
    dirbuf:prootdirs;
    p:pointer; {temp pointer}
    dirsize,w,slackstart,slack,numclusters,dirnum,clusval,clusnum:word;
    l:longint; {temp var}
    fname:string[12];
  begin
    fname[0]:=#12; fname[9]:='.';
    dirsize:=numentries*32;
    getmem(dirbuf,dirsize);
    if (dirsize mod 512)<>0 then abort(20,'# of directory entries not even multiple of sector size');
    DOSAbsDiskRead(DriveInfo.Letter,startsec,dirsize div 512,dirbuf,gIOerr);
    for dirnum:=0 to numentries-1 do begin
      {if no more entries, break out of loop}
      if dirbuf^[dirnum].name[0]=chr($00) then break;
      {if . or .. or erased then ignore}
      if not (dirbuf^[dirnum].name[0] in ['.',chr($e5)]) then begin
        {if a directory entry, follow it}
        if (dirbuf^[dirnum].filesize=0) and (dirbuf^[dirnum].attr AND da_DIR=da_DIR) then begin
          l:=firstSectorOfCluster(dirbuf^[dirnum].FirstClusLO);
          move(dirbuf^[dirnum].name,ptr(seg(fname),ofs(fname)+1)^,8);
          fname[10]:=dirbuf^[dirnum].ext[0]; fname[11]:=dirbuf^[dirnum].ext[1]; fname[12]:=dirbuf^[dirnum].ext[2];
          write(#13#10); for w:=0 to (level * 2) do write(#32);
          write(#192'>',fname);
          if verbose then writeln;
          IterateDIR(dirbuf^[dirnum].FirstClusLO,l,clusbytes div 32,level+1,printonly);
        end;
        {
        if filesize<>0 and filesize not exact multiple of cluster
        then we have some slack space after the end of the file}
        if (dirbuf^[dirnum].filesize<>0)
        and (dirbuf^[dirnum].filesize mod clusbytes<>0) then begin
          slackstart:=(dirbuf^[dirnum].filesize mod clusbytes);
          slack:=clusbytes-slackstart;
          {
          If not printonly, we must save the slack to a file. We will have to
          take the starting cluster and follow it until we have come to the
          end-of-file marker, then save our slack from that cluster.}
          if not printonly then begin
            numclusters:=1+(dirbuf^[dirnum].filesize div clusbytes);
            {if filesize mod clusbytes=0 then dec(numclusters), but we don't
            do that because we wouldn't be here if that were the case}
            {Avoid repeat..until loop because it will loop endlessly if FAT is
            damaged.  Instead, limit to exact number of iterations necessary.}
            clusnum:=dirbuf^[dirnum].FirstClusLO;
            clusval:=FATVal(clusnum);
            for w:=1 to numclusters-1 do begin {this gets skipped entirely if numclusters=1}
              if clusval>=DriveInfo.EOC then break; {in case our math is off-by-1}
              clusnum:=clusval;
              clusval:=FATVal(clusnum);
            end;
            if clusval>=DriveInfo.EOC then begin
              ReadCluster(clusnum,clusbuf);
              p:=clusbuf;
              inc(word(p),slackstart);
              blockwrite(f,p^,slack,gIOerr);
              if gIOerr<>slack
                then abort(13,'Error writing '+inttostr(slack)+' bytes to slackspace file ('+inttostr(gIOerr)+' written)');
              {throw user a bone if full verbosity is false}
              if not verbose then if (totalfiles AND $0F)=0 then write('.');
            end else abort(12,'End of cluster chain not EOC marker; possible FAT damage, run CHKDSK');
          end;
          inc(totalslack,slack);
          inc(totalfiles);
          if verbose then begin
            move(dirbuf^[dirnum].name,ptr(seg(fname),ofs(fname)+1)^,8);
            fname[10]:=dirbuf^[dirnum].ext[0]; fname[11]:=dirbuf^[dirnum].ext[1]; fname[12]:=dirbuf^[dirnum].ext[2];
            {$IFDEF DEBUG} write(#10); {$ENDIF}
            write(#13); for w:=0 to (level * 2) do write(#32);
            write(fname,' size: ',intpadded(dirbuf^[dirnum].filesize,10,#32),
              ' slack: ',intpadded(slack,5,#32),
              ' starting at: ',intpadded(slackstart,5,#32));
          end; {verboseprint}
        end; {foundslack}
      end; {validfilename}
    end; {numentries}
    {
    If we are here, it is because of the following possibilities:
    1. end of root directory entries (break) - we are done
    2. end of subdir entries (break) - we are done
    3. # of entries fills cluster exactly - we are done
    4. ran out of subdir entries for this cluster - chain to next cluster
    The only thing we need to check for is #4, by checking if its cluster
    chains to a value of >= $ff8/$fff8 -- if so, we're done, else iterate
    from there.  level>0 is subdir.}
    if (level>0) and (FATVal(startClus)<DriveInfo.EOC) then begin
      {chain to next subdirectory cluster}
      l:=firstSectorOfCluster(FATVal(startClus));
      if verbose then begin
        writeln;
        for w:=0 to (level * 2) do write(#32);
        writeln('Chaining to subdirectory cluster ',inttostr(FATVal(startclus)));
      end; {verboseprint}
      {iterate to next subdir chain; same numentries, level, and print status}
      IterateDIR(FATVal(startclus),l,numentries,level,printonly);
    end;

    freemem(dirbuf,dirsize);
  end; {IterateDIR}

begin
  totalslack:=0; totalfiles:=0;
  clusbytes:=(DriveInfo.SPC * 512);
  getmem(clusbuf,clusbytes); {allocate here so that out-of-memory occurs during scan phase and not write phase}
  writeln('Searching for file slack space:');
  IterateDIR(0,DriveInfo.RootStartSec,DPT.NRDE,0,true);
  {ask user if they want to save it to a file}
  if totalslack<>0 then begin
    writeln;
    writeln(#13#10'Found ',totalslack div 1024,' KiB total slack space in ',totalfiles,' files.');
    writeln(#13#10'*WARNING!* Do NOT save slack space to the SAME drive unless you have already');
    writeln('frozen the deleted data space (or don''t care about overwriting it).');
    if not confirm('Do you want save all drive '+DriveInfo.Letter+': file slack space to a new file?')
      then exit;
    write('(X:\PATH\FILENAME.EXT) Filename: ');
    readln(s);
    assign(f,s);
    rewrite(f,1);
    totalslack:=0; totalfiles:=0;
    IterateDIR(0,DriveInfo.RootStartSec,DPT.NRDE,0,false);
    close(f);
    writeln(#13#10'Saved ',totalslack div 1024,' KiB of file slack space into ',s);
  end;
  freemem(clusbuf,clusbytes);
end;

procedure FreezeDeleted;
var
  rootdir:prootdirs;
  rootdirsize,w,w2,freeEntry,firstfreeclus,lastfreeclus,clussize:word;
  clusbuf:pointer;

  function FindNextFreeDelClus(c:word):word;
  {Finds the next free deleted cluster based on whether or not it contains}
  {only the formatted $f6 value (or any single value if aggresiveFsearch=true).}
  var
    tmpc:word;
  begin
    tmpc:=c;
    repeat
      tmpc:=FindNextFreeClus(tmpc);
      if tmpc=0 then break; {no more free clusters to check}
      if verbose
        then write(#13'Checking ',numfmt(tmpc))
        else if (tmpc and $ff)=$ff
          then write(#13,inttostr((longint(tmpc) * 100) div longint(DriveInfo.NC)),'%');
      ReadCluster(tmpc,clusbuf);
      if not SameVal(clusbuf,clussize,$f6) then begin
        if verbose then write(' Found at ',numfmt(tmpc));
        break; {we found clus with deleted data}
      end;
      inc(tmpc);
    until tmpc>DriveInfo.NC+1; {should never get here, but just in case}
    if tmpc>DriveInfo.NC+1 then tmpc:=0;
    FindNextFreeDelClus:=tmpc;
  end;

begin
  writeln(#13#10'!!WARNING!!');
  writeln('"Freezing" all of the deleted data on drive ',DriveInfo.Letter,': is a non-destructive operation,');
  writeln('but it does rewrite the root directory as well as the first copy of the FAT.');
  writeln('This has the potential to damage the filesystem if something goes wrong while');
  writeln('operations are taking place, such as a power outage.  Proceed at your own risk.');
  if not confirm('Do you wish to continue?')
    then exit;
  clussize:=512*DriveInfo.SPC;
  getmem(clusbuf,clussize);

  {load root directory}
  writeln('Loading ',DPT.NRDE,' root directory entries...');
  rootdirsize:=DPT.NRDE*32;
  getmem(rootdir,rootdirsize);
  if (rootdirsize mod 512)<>0 then abort(20,'# of directory entries not even multiple of sector size');
  DOSAbsDiskRead(DriveInfo.Letter,DriveInfo.RootStartSec,rootdirsize div 512,rootdir,gIOerr);

  {find unused entry for us to use; if none, use erased entry; if none, abort}
  write('Searching for unused directory entry we can use... ');
  freeEntry:=$ffff;
  for w:=0 to DPT.NRDE-1 do
    if rootdir^[w].name[0]=chr($00) {unused entry} then begin
      freeEntry:=w;
      break;
    end;
  if freeEntry=$ffff then begin
    write(#13'Couldn''t find an unused entry; searching for an erased entry... ');
    for w:=0 to DPT.NRDE-1 do
      if rootdir^[w].name[0]=chr($E5) {erased entry} then begin
        freeEntry:=w;
        break;
      end;
  end;
  if freeEntry=$ffff then abort(21,'Root dir full!  Delete a file or the volume label to resolve.');
  writeln(freeEntry);

  {find first free cluster not sameval and record into directory entry}
  writeln('Examining all free clusters for deleted data...');
  w:=FindNextFreeDelClus(2); if w=0 then abort(22,'No deleted clusters found; nothing to freeze.');
  firstfreeclus:=w;
  with rootdir^[freeEntry] do begin
    FirstClusLO:=firstfreeclus;
    filesize:=clussize;
    name[0]:='D';name[1]:='E';name[2]:='L';name[3]:='E';name[4]:='T';name[5]:='E';name[6]:='D';name[7]:=#32;
    ext[0]:='D';ext[1]:='A';ext[2]:='T';
    {make up a date.  I choose the birth of the IBM PC:}
    asm
      {august 12th, 1981}
      mov ax,0000001100001100b
      mov w2,ax
    end;
    WriteDate:=w2;
    (*WriteTime:=WriteDate; {just so it isn't 12:00am}*)
  end;
  {if disk has a volume label, steal the time/date from that}
  for w2:=0 to DPT.NRDE-1 do begin
    if (rootdir^[w2].attr and da_VOL)=da_VOL then begin
      rootdir^[freeEntry].WriteDate:=rootdir^[w2].WriteDate;
      rootdir^[freeEntry].WriteTime:=rootdir^[w2].WriteTime;
      break;
    end;
  end;
  {find rest of free clusters not sameval and create cluster chain in fatcopy}
  lastfreeclus:=firstfreeclus;
  repeat
    w:=FindNextFreeDelClus(w+1);
    if w<>0 then begin
      {point last cluster to new cluster}
      setFAT(lastfreeclus,w);
      {$IFDEF DEBUG} fatval(lastfreeclus); {$ENDIF}
      lastfreeclus:=w;
      inc(rootdir^[freeEntry].filesize,clussize);
    end;
  until w=0;
  {mark last cluster EOC and record total space as filesize}
  setFAT(lastfreeclus,DriveInfo.EOC);

  {write root directory back to disk}
  writeln(#13#10'Updating root directory...');
  DOSAbsDiskWrite(DriveInfo.Letter,DriveInfo.RootStartSec,rootdirsize div 512,rootdir,gIOerr);

  {write FAT back to disk}
  writeln('Flushing FAT cache...');
  FlushFATToDisk;

  {clean up}
  writeln('Found ',rootdir^[freeEntry].filesize,
    ' bytes of deleted data and froze them into ',DriveInfo.Letter,':\DELETED.DAT');
  freemem(rootdir,rootdirsize);
  freemem(clusbuf,clussize);
end;



end.