unit paddvars;

interface

type
  FATtype=(FAT12,FAT16,FAT32);
  Viewtype=(rawview,printable,hexview16,hexview20,hexview24);
  DriveType=record
    Letter:char;
    MediaDescriptor:byte;
    Removable:boolean;
    SPC:byte;             {sectors per cluster}
    BPS:word;             {bytes per sector}
    NC:word;              {number of clusters on disk}
    NFC:word;             {number of free clusters (from DOS)}
    CFC:word;             {number of free clusters we counted}
    FATString:string[5];  {either FAT12 or FAT16}
    FAT:FATtype;
    TotalBytes,
    FreeBytes:longint;
    FATstartsec:word;     {sector where 1st FAT starts}
    FATNumSectors:word;
    FATSizeb:longint;     {FAT size in butes}
    ROOTstartsec:word;    {sector where root directory starts}
    CLUSstartsec:word;    {sector where the first cluster starts}
    EOC:word;             {holds end-of-cluster-chain value}
  end;

  DiskParamTable=record
    drivenum  :byte;    {drive (0 = A, 1 = B)}
    driveunit :byte;    {unit within device, usually equals drive (ramdisk=0)}
    BPS       :word;    {bytes per sector}
    SPCM1     :byte;    {sectors per cluster minus 1}
    CLSShift  :byte;    {sectors per cluster (times to shift left or x2)}
    NSBFAT    :word;    {number of sectors before FAT (boot sectors)}
    FATcopies :byte;    {number of FAT copies}
    NRDE      :word;    {number of root directory entries}
    NFDS      :word;    {number of first data sector}
    TNCP1     :word;    {total number of clusters plus 1}
(*  DOS 3.30 and earlier: *)
(*  FATNumSecs:byte;    {number of sectors used by first FAT} {if 0 then =256}*)
(*  DOS 3.31 and later:   *)
    FATNumSecs:word;    {number of sectors used by first FAT}
    {We are going to read this as a WORD and on DOS versions or
    FAT12 filesystems where it is just a byte, we will mask out the
    upper byte which is junk in those circumstances.}
  end;

  RootEntry=record
    Name:array[0..7] of char;         {0}
    Ext:array[0..2] of char;          {8}
    Attr:byte;                        {11}
    NTRes:byte;                       {12}
    CrtTimeTenth:byte;                {13}
    CrtTime:word;                     {14}
    CrtDate:word;                     {16}
    LastAccDate:word;                 {18}
    FirstClusHI:word;                 {20}
    WriteTime:word;                   {22}
    WriteDate:word;                   {24}
    FirstClusLO:word;                 {26}
    FileSize:longint;                 {28}
  end;

  PRootdirs=^Trootdirs;
  Trootdirs=array[0..(65520 div sizeof(RootEntry))-1] of RootEntry;

  PCharArray=^TCharArray;
  TCharArray=array[0..65520-1] of char;

  PByteArray=^TCharArray;
  TByteArray=array[0..65520-1] of byte;

  PWordArray=^TCharArray;
  TWordArray=array[0..32760-1] of word;

const
  progbanner='PADD - Peek At Deleted Data | Version 1.0 | Jim Leonard, trixter@oldskool.org';
  DPTsize=sizeof(DiskParamTable);
  viewmode:viewtype=rawview;
  numhex:boolean=true;       {number formats printed in hexidecimal}
  spinners:array[0..3] of char = ('','\','','/');

  {command-line args}
  verbose:boolean=false;
  skipprompts:boolean=false;
  aggressiveFsearch:boolean=false;
  {aggresive format search.  if true, will test free clusters for 100% fill
  of *any* value, not just $f6}

  {directory attribute masks}
  da_RO =$01;
  da_HID=$02;
  da_SYS=$04;
  da_VOL=$08;
  da_DIR=$10;
  da_ARC=$20;
  da_LFN=(da_RO OR da_HID OR da_SYS OR da_VOL);

  {volume clean masks}
  FAT16_ClnShutdown=$8000;
  FAT16_NoHardError=$4000;

  {FAT memory handling}
  segsize=$8000;
  secsPerSeg=segsize div 512;

var
  DriveInfo:Drivetype;
  DPT:DiskParamTable;
  FATcopy,f1,f2,f3,f4:pointer;
  FATSegs:array[0..1] of word;
  gIOerr:word; {global IOError}

implementation

begin
end.